/*****************************************************************************
* Copyright (c) 1998-2004 Sena Technologies, Inc., All Rights Reserved
*
* This software contains proprietary and confidential information of Sena
* Technologies Inc.  By accepting transfer of this copy, Recipient agrees
* to retain this software in confidence, to prevent disclosure to others,
* and to make no use of this software other than that for which it was
* delivered.  This is an unpublished copyrighted work of Sena Technologies
* Inc.
*
* Sena Technologies, Inc. 210 Yangjae-dong, Seocho-gu Seoul 137-130, Korea
*
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "cq.h"

char* cq_fw_ptr (CQ *cq, char *p, size_t offset)
{
	offset %= cq->bufsize;
	p += offset;
	if (p >= cq->buf + cq->bufsize)
		p -= cq->bufsize;
	return p;
}
int cq_inuse (CQ *cq)
{
	return cq->bufsize - cq_left (cq);
}
int cq_left (CQ *cq)
{
	if (cq->frontp <= cq->rearp) 
		return cq->bufsize - (cq->rearp - cq->frontp);
	else
		return cq->frontp - cq->rearp;
}
int cq_shrink (CQ *cq, size_t len)
{
	if (cq_inuse (cq) < len) return -1;
	cq->frontp = cq_fw_ptr (cq, cq->frontp, len);
	return 0;
}
int cq_expand (CQ *cq, size_t len)
{
	if (cq_left (cq) < len) return -1;
	cq->rearp = cq_fw_ptr (cq, cq->rearp, len);
	return 0;
}
void cq_init (CQ *cq, size_t bufsize)
{
	cq->buf = (char*)malloc(bufsize);
	cq->frontp = cq->rearp = cq->buf;
	cq->bufsize = bufsize;
}
void cq_free(CQ *cq)
{
	free(cq->buf);
}

void cq_reset (CQ *cq)
{
	cq->rearp = cq->buf;
	cq->frontp = cq->buf;
}

int cq_read (CQ *cq, char *from, char *buf, size_t len)
{
	size_t chunk;

	chunk = (size_t)cq->buf + cq->bufsize - (size_t)from;
	if (chunk >= len)
	{
		memcpy(buf, from, len);
	}
	else
	{
		memcpy(buf, from, chunk);
		memcpy(buf + chunk, cq->buf, len - chunk);
	}
	return 0;
}

int cq_write (CQ *cq, char *to, char *buf, size_t len)
{
	size_t chunk;

	chunk = (size_t)cq->buf + cq->bufsize - (size_t)to;
	if (chunk >= len)
	{
		memcpy(to, buf, len);
	}
	else
	{
		memcpy(to, buf, chunk);
		memcpy(cq->buf, buf + chunk, len - chunk);
	}

	return 0;
}


int cq_getch(CQ *cq)
{
	int ch = -1;
	if (cq_inuse (cq)>0) {
		ch = (int)*(cq->frontp); 
		cq_shrink (cq, 1);
	}
	return ch;
}

int cq_putch(CQ *cq, char ch)
{
	int nret = -1;
	if (cq_left (cq)>0) {
		*(cq->rearp) = ch; 
		cq_expand (cq, 1);
		nret = 0;
	}
	return nret;
}

