/*****************************************************************************
* Copyright (c) 1998-2004 Sena Technologies, Inc., All Rights Reserved
* 
* This software contains proprietary and confidential information of Sena
* Technologies Inc.  By accepting transfer of this copy, Recipient agrees
* to retain this software in confidence, to prevent disclosure to others,
* and to make no use of this software other than that for which it was
* delivered.  This is an unpublished copyrighted work of Sena Technologies
* Inc.   
*
* Sena Technologies, Inc. 210 Yangjae-dong, Seocho-gu Seoul 137-130, Korea
*
*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>

#include "cq.h" // circular queue manipulcation functions

#define BUFFER_SIZE 	4096
#define PID_FILE	"/var/run/port%d_filter.pid"
#define STOF_FIFO	"/tmp/port_fifos/port%d_s2f"
#define FTOE_FIFO	"/tmp/port_fifos/port%d_f2e"
#define ETOF_FIFO	"/tmp/port_fifos/port%d_e2f"
#define FTOS_FIFO	"/tmp/port_fifos/port%d_f2s"

#define NOTIFICATION_INTERVAL		10
#define MAX_NUM_VALUES			10		

#define MAX_PACKET_SIZE			4  //STX(1 byte)+DATA(2 bytes)+ETX(1 byte)

#define _STX	0x02
#define _ETX	0x03
#define _CR	'\r'
#define _ACK	0x06
#define _NAK	0x15


int s2f_fd = -1;
int f2s_fd = -1;
int e2f_fd = -1;
int f2e_fd = -1;

int exit_flag = 0;
int portnum = -1;


unsigned short int ave;

CQ serial_cq;

void do_filter();
int open_fifos(int port);
void close_fifos();

void handle_sigterm(int sig)
{
	exit_flag = 1;
}

int do_daemon()
{
	int pid;
	
	switch (pid=fork())
	{
		case -1: return -1;
		case 0: break;
		default: _exit(0);
	}
	if (setsid() == -1) return -1;

}

//
// create a file(/var/run/portX_filter.pid) to save process ID	
//
void save_pid(int port)
{
	FILE *fp;
	char filename[128];

	sprintf(filename, PID_FILE, port);
	
	fp = fopen(filename, "w");
	
	if (fp) {
		fprintf(fp, "%d", getpid());
		fclose(fp);
	}
}


//
// get program name from the first argument
//
char *get_program_name(char *str)
{
	char *p = strrchr(str, '/');

	if (p) return p+1;
	return str;	
}


//
// main function
//
int main(int argc, char **argv)
{
	if (argc != 2 )	{
		fprintf(stdout, "\nUsage: %s [portnumber]\n\n", get_program_name(argv[0]));
		return -1;
	}	
	
	portnum = atoi(argv[1]); 
		
	(void) signal(SIGTERM, handle_sigterm);
	(void) signal(SIGPIPE, handle_sigterm);

	do_daemon();
	save_pid(portnum);	
	
	do_filter();
	close_fifos();

	return 0;
}

//
// close all fifos
//
void close_fifos()
{
	close(s2f_fd);
	close(f2s_fd);
	close(e2f_fd);
	close(f2e_fd);
}

//
// calculate average value
//
unsigned short int average(short int value[], int num_of_values)
{
	unsigned int sum = 0, i;
	
	for (i=0; i<num_of_values; i++) {
		sum += (int) value[i];
	}
	return ((short int) sum/num_of_values);
}	

//
// read on serial port, calculate average of latest data   
//
// serial device protocol: 
// 	STX + data(16bits integer) + ETX
//
void *read_serial_thread (void *arg)
{
	char buf[BUFFER_SIZE], *to;
	int nread=0, num_of_val=0, next_val=0, total;
	char fifoname[128];
	unsigned short int values[MAX_NUM_VALUES], temp;

	
	// enable cancelation and set cancelation type to asynchronous 	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, STOF_FIFO, portnum);
	s2f_fd = open(fifoname, O_RDONLY);
	if (s2f_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		// read on serial port
		nread = read(s2f_fd, buf, sizeof(buf));
		if (nread<=0) continue;

		// put the received data to circular buffer
		to = serial_cq.rearp;
		if (!cq_expand(&serial_cq, nread)) cq_write(&serial_cq, to, buf, nread);	
		
		// read a character form circular buffer and handle it
		while (cq_inuse(&serial_cq)>=MAX_PACKET_SIZE) {
			if (cq_getch(&serial_cq) == _STX) {	
				temp = cq_getch(&serial_cq) << 8;	
				temp |= cq_getch(&serial_cq);	
				if (cq_getch(&serial_cq) == _ETX) {	
					values[next_val] = temp;
					if (num_of_val < MAX_NUM_VALUES) num_of_val++;
					if (++next_val == MAX_NUM_VALUES) next_val=0;
					ave = average(values, num_of_val); 
				}	
			}
		}
	}
	pthread_exit(NULL);
}

//
// send the average to socket
//
void *write_socket_thread (void *arg)
{
	char buf[128];
	char fifoname[128];

	
	// enable cancelation and set cancelation type to asynchronous 	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	
	sprintf(fifoname, FTOE_FIFO, portnum);
	f2e_fd = open(fifoname, O_WRONLY);
	if (f2e_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		sleep(NOTIFICATION_INTERVAL);
		sprintf(buf, "%4.4d\r", ave);
		write(f2e_fd, buf, strlen(buf));
	}
	pthread_exit(NULL);
}


//
// read on socket, but do nothing.
//
void *e2s_thread (void *arg)
{
	char fifoname[128];

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, ETOF_FIFO, portnum);
	e2f_fd = open(fifoname, O_RDONLY);
	if (e2f_fd < 0) pthread_exit(NULL);
	
	sprintf(fifoname, FTOS_FIFO, portnum);
	f2s_fd = open(fifoname, O_WRONLY);
	if (f2s_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		// do nothing.
		sleep(100);
	}
	pthread_exit(NULL);
}




void do_filter()
{
	int result;
	pthread_t read_serial_th, write_socket_th,e2s_th;
	void *thread_result;

	// initialize circular buffer
	cq_init(&serial_cq, BUFFER_SIZE);	
	
	// create threads	
	result = pthread_create(&read_serial_th, NULL, read_serial_thread, NULL);
	result = pthread_create(&write_socket_th, NULL, write_socket_thread, NULL);
	result = pthread_create(&e2s_th, NULL, e2s_thread, NULL);

	// waitl until reeceives SIGTERM	
	while(!exit_flag) pause();
	
	// cancel all threads	
	pthread_cancel(read_serial_th);
	pthread_cancel(write_socket_th);
	pthread_cancel(e2s_th);
	
	// wait until all threads is canceled.	
	pthread_join(read_serial_th, &thread_result);
	pthread_join(write_socket_th, &thread_result);
	pthread_join(e2s_th, &thread_result);

	// free allocated memory for circular buffer	
	cq_free(&serial_cq);	
}


