/*****************************************************************************
* Copyright (c) 1998-2004 Sena Technologies, Inc., All Rights Reserved
* 
* This software contains proprietary and confidential information of Sena
* Technologies Inc.  By accepting transfer of this copy, Recipient agrees
* to retain this software in confidence, to prevent disclosure to others,
* and to make no use of this software other than that for which it was
* delivered.  This is an unpublished copyrighted work of Sena Technologies
* Inc.   
*
* Sena Technologies, Inc. 210 Yangjae-dong, Seocho-gu Seoul 137-130, Korea
*
*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>

#include "cq.h" // circular queue manipulcation functions

#define BUFFER_SIZE 	4096
#define BUFFER_SIZE 	4096

#define PID_FILE	"/var/run/port%d_filter.pid"
#define STOF_FIFO	"/tmp/port_fifos/port%d_s2f"
#define FTOE_FIFO	"/tmp/port_fifos/port%d_f2e"
#define ETOF_FIFO	"/tmp/port_fifos/port%d_e2f"
#define FTOS_FIFO	"/tmp/port_fifos/port%d_f2s"

#define _STX	0x02
#define _ETX	0x03
#define _CR	'\r'
#define _ACK	0x06
#define _NAK	0x15

#define DELIMITER	_CR


#define MAX_SERIAL_PACKET_SIZE	1024



int s2f_fd = -1;
int f2s_fd = -1;
int e2f_fd = -1;
int f2e_fd = -1;

int exit_flag = 0;
int portnum = -1;

int echo_flag = 1;


pthread_mutex_t socket_write_mutex;

CQ s2e_cq, e2s_cq;

void do_filter();
int open_fifos(int port);
void close_fifos();

void handle_sigterm(int sig)
{
	exit_flag = 1;
}

int do_daemon()
{
	int pid;
	
	switch (pid=fork())
	{
		case -1: return -1;
		case 0: break;
		default: _exit(0);
	}
	if (setsid() == -1) return -1;

}

//
// create a file(/var/run/portX_filter.pid) to save process ID	
//
void save_pid(int port)
{
	FILE *fp;
	char filename[128];

	sprintf(filename, PID_FILE, port);
	
	fp = fopen(filename, "w");
	
	if (fp) {
		fprintf(fp, "%d", getpid());
		fclose(fp);
	}
}

//
// get program name from the first argument
//
char *get_program_name(char *str)
{
	char *p = strrchr(str, '/');

	if (p) return p+1;
	return str;	
}

//
// main function
//
int main(int argc, char **argv)
{
	if (argc < 2)	{
		fprintf(stderr, "\nUsage: %s [portnumber] [echo|no_echo]\n\n", get_program_name(argv[0]));
		return -1;
	}	
	
	portnum = atoi(argv[1]);
	if (argc>2 && !strcmp(argv[2], "no_echo")) echo_flag = 0;
		
	(void) signal(SIGTERM, handle_sigterm);
	(void) signal(SIGPIPE, handle_sigterm);

	do_daemon();
	save_pid(portnum);	
	
	do_filter();
	close_fifos();

	return 0;
}

//
// close all fifos	
//
void close_fifos()
{
	close(s2f_fd);
	close(f2s_fd);
	close(e2f_fd);
	close(f2e_fd);
}


//
// read on serial port, convert the received data and write the converted data to socket
// 
// conversion: 
// 	STX+ ACK or NAK + ETX -> "OK" or "NG" +CR
//
// if receives other formed data, ignore it.		 	 
//
void *s2e_thread (void *arg)
{
	const char szOK[] = "OK\r";
	const char szNG[] = "NG\r";
	const char szERROR[] = "ERROR\r";
	
	char buf[BUFFER_SIZE], ch;
	int nread=0;
	char fifoname[128], *to, *from, *current;

	// enable cancelation and set cancelation type to asynchronous 	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, STOF_FIFO, portnum);
	s2f_fd = open(fifoname, O_RDONLY);
	if (s2f_fd < 0) pthread_exit(NULL);
	
	sprintf(fifoname, FTOE_FIFO, portnum);
	f2e_fd = open(fifoname, O_WRONLY);
	if (f2e_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		// read on serial port
		nread = read(s2f_fd, buf, sizeof(buf));
		if (nread<=0) continue;

		// put the received data to circular buffer
		to = s2e_cq.rearp;
		if (!cq_expand(&s2e_cq, nread)) cq_write(&s2e_cq, to, buf, nread);	
		
		while(cq_inuse(&s2e_cq)) {
			// read a character form circular buffer
			from = s2e_cq.frontp;
			cq_read(&s2e_cq, from, &ch, 1); 	
			// if the character is STX and the next character is ACK or NAK write "OK\r" or "NG\r" to socket
			// else ignore it 
			if (ch == _STX) {
				if (cq_inuse(&s2e_cq)<3) break;
				cq_shrink(&s2e_cq, 1);	
				from = s2e_cq.frontp;
				cq_read(&s2e_cq, from, buf, 2); 	
				cq_shrink(&s2e_cq, 2);	
				if (buf[1] == _ETX) {
					pthread_mutex_lock(&socket_write_mutex);
					if (buf[0] == _ACK) write(f2e_fd, szOK, strlen(szOK));
					else if (buf[0] == _NAK) write(f2e_fd, szNG, strlen(szNG));
					else write(f2e_fd, szERROR, strlen(szERROR));
					pthread_mutex_unlock(&socket_write_mutex);
				}	
			} else {
				// remove the handled data in circular buffer
				cq_shrink(&s2e_cq, 1);	
			}		
		}
	}
	pthread_exit(NULL);
}

//
// form data by serial device protocol(STX + data+ ETX) and  send it to serial port 
//
int send_to_serial_port(int fd, char *buf, int buf_len)
{
	char send_buf[BUFFER_SIZE];
	char temp[128];
	int nbytes_to_send = buf_len+2;

	
	if (buf_len <= 0) return -1;
		
	send_buf[0] = _STX;
	memcpy(send_buf+1, buf, buf_len);
	send_buf[buf_len+1] = _ETX;

	memset(temp, 0x00, sizeof(temp));
	memcpy(temp, buf, buf_len);
	
	
	if (write(fd, send_buf, nbytes_to_send)!= nbytes_to_send) return -1;	
	return 0;
}	


//
// read on socket, convert the received data and write the converted data to serial port
// 
// conversion: 
// 	data(string) + CR -> STX + data(string) + ETX
//
// if receives other formed data, ignore it.		 	 
//
void *e2s_thread (void *arg)
{
	char buf[BUFFER_SIZE];
	int nread=0;
	char fifoname[128], *to, *from, *start, *current;

	// enable cancelation and set cancelation type to asynchronous 	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, ETOF_FIFO, portnum);
	e2f_fd = open(fifoname, O_RDONLY);
	if (e2f_fd < 0) pthread_exit(NULL);
	
	sprintf(fifoname, FTOS_FIFO, portnum);
	f2s_fd = open(fifoname, O_WRONLY);
	if (f2s_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		// read on socket
		nread = read(e2f_fd, buf, sizeof(buf));
		if (nread<=0) continue;
		// if TCP echo is enabled, send echo
		if (echo_flag) {
			pthread_mutex_lock(&socket_write_mutex);
			if (f2e_fd>0) write(f2e_fd, buf, nread);
			pthread_mutex_unlock(&socket_write_mutex);
		}
		
		// put the received data to circular buffer
		to = e2s_cq.rearp;
		if (!cq_expand(&e2s_cq, nread)) cq_write(&e2s_cq, to, buf, nread);	
	
		// read data from circular buffer
		nread = cq_inuse(&e2s_cq);
		from = e2s_cq.frontp;
		cq_read(&e2s_cq, from, buf, nread); 	
		
		start = current = buf;
		// find the delimiter(=CR) and send the ahead data of delimiter to serial port
		for(;current<buf+nread; current++) { 	
			if(*current == DELIMITER) {
				send_to_serial_port(f2s_fd, start, current-start);
				// remove the handled data in circular buffer
				cq_shrink(&e2s_cq, current-start+1);	
				start = current+1;
			}
		}	
		
		// flush the circular buffer, in order to prevent buffer overflow 
		if (cq_inuse(&e2s_cq) > MAX_SERIAL_PACKET_SIZE) cq_reset(&e2s_cq);
		
	}
	pthread_exit(NULL);
}




void do_filter()
{
	int result;
	pthread_t s2e_th, e2s_th;
	void *thread_result;

	// initialize circular buffers
	cq_init(&s2e_cq, BUFFER_SIZE);	
	cq_init(&e2s_cq, BUFFER_SIZE);	

	result = pthread_mutex_init(&socket_write_mutex, NULL);
	if (result) return ;
	
	// create threads	
	result = pthread_create(&s2e_th, NULL, s2e_thread, NULL);
	result = pthread_create(&e2s_th, NULL, e2s_thread, NULL);

	// waitl until reeceives SIGTERM	
	while(!exit_flag) pause();
	
	// cancel all threads	
	pthread_cancel(s2e_th);
	pthread_cancel(e2s_th);
	
	// wait until all threads is canceled.	
	pthread_join(s2e_th, &thread_result);
	pthread_join(e2s_th, &thread_result);

	pthread_mutex_destroy(&socket_write_mutex);

	// free allocated memory for circular buffers	
	cq_free(&s2e_cq);	
	cq_free(&e2s_cq);	
}


