/*****************************************************************************
* Copyright (c) 1998-2004 Sena Technologies, Inc., All Rights Reserved
* 
* This software contains proprietary and confidential information of Sena
* Technologies Inc.  By accepting transfer of this copy, Recipient agrees
* to retain this software in confidence, to prevent disclosure to others,
* and to make no use of this software other than that for which it was
* delivered.  This is an unpublished copyrighted work of Sena Technologies
* Inc.   
*
* Sena Technologies, Inc. 210 Yangjae-dong, Seocho-gu Seoul 137-130, Korea
*
*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>
#include <time.h>

#include "cq.h" // circular queue manipulcation functions

#define BUFFER_SIZE 	4096
#define PID_FILE	"/var/run/port%d_filter.pid"
#define STOF_FIFO	"/tmp/port_fifos/port%d_s2f"
#define FTOE_FIFO	"/tmp/port_fifos/port%d_f2e"
#define ETOF_FIFO	"/tmp/port_fifos/port%d_e2f"
#define FTOS_FIFO	"/tmp/port_fifos/port%d_f2s"

#define MAX_VALUE		 	100				
#define MAX_PACKET_SIZE			4  //STX(1 byte)+DATA(2 bytes)+ETX(1 byte)

#define _STX	0x02
#define _ETX	0x03
#define _CR	'\r'
#define _ACK	0x06
#define _NAK	0x15

// SNMP variables, modify it to yours.
#define ENTERPRISE_OID	".1.3.6.1.4.1.12236"
#define SNMPTRAP_OID 	".1.3.6.1.4.1.12236.100"
#define TARGET_IP    	"192.168.11.150"
#define COMMUNITY	"public"

int s2f_fd = -1;
int f2s_fd = -1;
int e2f_fd = -1;
int f2e_fd = -1;

int exit_flag = 0;
int portnum = -1;


CQ serial_cq;

void do_filter();
int open_fifos(int port);
void close_fifos();

void handle_sigterm(int sig)
{
	exit_flag = 1;
}

int do_daemon()
{
	int pid;
	
	switch (pid=fork())
	{
		case -1: return -1;
		case 0: break;
		default: _exit(0);
	}
	if (setsid() == -1) return -1;

}

//
// create a file(/var/run/portX_filter.pid) to save process ID	
//
void save_pid(int port)
{
	FILE *fp;
	char filename[128];

	sprintf(filename, PID_FILE, port);
	
	fp = fopen(filename, "w");
	
	if (fp) {
		fprintf(fp, "%d", getpid());
		fclose(fp);
	}
}

	
//
// get program name from the first argument
//
char *get_program_name(char *str)
{
	char *p = strrchr(str, '/');

	if (p) return p+1;
	return str;	
}


//
// main function
//
int main(int argc, char **argv)
{
	if (argc != 2 )	{
		fprintf(stdout, "\nUsage: %s [portnumber]\n\n", get_program_name(argv[0]));
		return -1;
	}	
	
	portnum = atoi(argv[1]); 
		
	(void) signal(SIGTERM, handle_sigterm);
	(void) signal(SIGPIPE, handle_sigterm);

	do_daemon();
	save_pid(portnum);	
	
	do_filter();
	close_fifos();

	return 0;
}

//
// close all fifos
//
void close_fifos()
{
	close(s2f_fd);
	close(f2s_fd);
	close(e2f_fd);
	close(f2e_fd);
}

//
// send snmptrap
//
void send_snmptrap(unsigned short value)
{
	char szcommand[512], sztime[128], szmessage[256];
        time_t timeval ;
        struct tm *tp ;

	time(&timeval) ;
	tp = (struct tm *)gmtime(&timeval);
	strftime(sztime, sizeof(sztime), "Date: %a , %d %b %Y %H:%M:%S", tp);
	sprintf(szmessage, "%s > %d", sztime, value);
	
	snprintf(szcommand, sizeof(szcommand), "/usr/local/sbin/snmptrap -v 1 %s %s "ENTERPRISE_OID" %s %u %u %u "
			SNMPTRAP_OID" s \"%s\"",
			TARGET_IP,       // ip
			COMMUNITY, // community
			"localhost", //ip address
			6, // generic trap
			1, // specific trap
			0, // uptime
			szmessage //trap message
			) ;
	
	system(szcommand);

}

//
// read on serial port and if the value is more than MAX_VALUE, send snmptrap. 
//
// serial device protocol: 
// 	STX + data(16bits integer) + ETX
//
void *read_serial_thread (void *arg)
{
	char buf[BUFFER_SIZE], *to;
	int nread=0;
	char fifoname[128];
	unsigned short value;

	
	// enable cancelation and set cancelation type to asynchronous 	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, STOF_FIFO, portnum);
	s2f_fd = open(fifoname, O_RDONLY);
	if (s2f_fd < 0) pthread_exit(NULL);

	sprintf(fifoname, FTOE_FIFO, portnum);
	f2e_fd = open(fifoname, O_WRONLY);
	if (f2e_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		// read on serial port
		nread = read(s2f_fd, buf, sizeof(buf));
		if (nread<=0) continue;

		// put the received data to circular buffer
		to = serial_cq.rearp;
		if (!cq_expand(&serial_cq, nread)) cq_write(&serial_cq, to, buf, nread);	
		
		// read a character form circular buffer and handle it
		while (cq_inuse(&serial_cq)>=MAX_PACKET_SIZE) {
			if (cq_getch(&serial_cq) == _STX) {	
				value = cq_getch(&serial_cq) << 8;	
				value |= cq_getch(&serial_cq);	
				if (cq_getch(&serial_cq) == _ETX) {
					if (value>=MAX_VALUE) send_snmptrap(value); 
				}	
			}
		}
	}
	pthread_exit(NULL);
}

//
// every 10 seconds, send STX+"REQUEST"+ETX to serial port in order to receive data from serial port continuosly.
// opens fifo related to socket application, but do noting to it.
//
void *e2s_thread (void *arg)
{
	const char szREQUEST[] = "\x02""REQUEST""\x03";
	char fifoname[128];

	// every 10 seconds, send STX+"REQUEST"+ETX to serial port
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, ETOF_FIFO, portnum);
	e2f_fd = open(fifoname, O_RDONLY);
	if (e2f_fd < 0) pthread_exit(NULL);
	
	sprintf(fifoname, FTOS_FIFO, portnum);
	f2s_fd = open(fifoname, O_WRONLY);
	if (f2s_fd < 0) pthread_exit(NULL);
	
	while(!exit_flag) {
		sleep(10);
		write(f2s_fd, szREQUEST, strlen(szREQUEST));	
	}
	pthread_exit(NULL);
}




void do_filter()
{
	int result;
	pthread_t read_serial_th, e2s_th;
	void *thread_result;

	// initialize circular buffer
	cq_init(&serial_cq, BUFFER_SIZE);	
	
	// create threads	
	result = pthread_create(&read_serial_th, NULL, read_serial_thread, NULL);
	result = pthread_create(&e2s_th, NULL, e2s_thread, NULL);

	// waitl until reeceives SIGTERM	
	while(!exit_flag) pause();
	
	// cancel all threads	
	pthread_cancel(read_serial_th);
	pthread_cancel(e2s_th);
	
	// wait until all threads is canceled.	
	pthread_join(read_serial_th, &thread_result);
	pthread_join(e2s_th, &thread_result);

	// free allocated memory for circular buffer	
	cq_free(&serial_cq);	
}


