/*****************************************************************************
* Copyright (c) 1998-2004 Sena Technologies, Inc., All Rights Reserved
* 
* This software contains proprietary and confidential information of Sena
* Technologies Inc.  By accepting transfer of this copy, Recipient agrees
* to retain this software in confidence, to prevent disclosure to others,
* and to make no use of this software other than that for which it was
* delivered.  This is an unpublished copyrighted work of Sena Technologies
* Inc.   
*
* Sena Technologies, Inc. 210 Yangjae-dong, Seocho-gu Seoul 137-130, Korea
*
*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <pthread.h>

#define BUFFER_SIZE 	4096
#define PID_FILE	"/var/run/port%d_filter.pid"
#define STOF_FIFO	"/tmp/port_fifos/port%d_s2f"
#define FTOE_FIFO	"/tmp/port_fifos/port%d_f2e"
#define ETOF_FIFO	"/tmp/port_fifos/port%d_e2f"
#define FTOS_FIFO	"/tmp/port_fifos/port%d_f2s"

#define PERIOD_DEFAULT	30

int s2f_fd = -1;
int f2s_fd = -1;
int e2f_fd = -1;
int f2e_fd = -1;

int bexit = 0;
int portnum = -1;
pthread_mutex_t serial_write_mutex;

char periodic_command[128] = "command";
int period_time = PERIOD_DEFAULT;

void do_filter();
int open_fifos(int port);
void close_fifos();

void handle_sigterm(int sig)
{
	bexit = 1;
}

int do_daemon()
{
	int pid;
	
	switch (pid=fork())
	{
		case -1: return -1;
		case 0: break;
		default: _exit(0);
	}
	if (setsid() == -1) return -1;

}

//
// create a file(/var/run/portX_filter.pid) to save process ID	
//
void save_pid(int port)
{
	FILE *fp;
	char filename[128];

	sprintf(filename, PID_FILE, port);
	
	fp = fopen(filename, "w");
	
	if (fp) {
		fprintf(fp, "%d", getpid());
		fclose(fp);
	}
}


//
// main function
//
int main(int argc, char **argv)
{
	if (argc != 2 && argc != 3 && argc !=4)	{
		fprintf(stdout, "\nUsage: %s [portnumber] [command] [period]\n\n", argv[0]);
		return -1;
	}
	
	portnum = atoi(argv[1]); 
	
	if (argc == 3 || argc == 4) {
		if (strlen(argv[2])) sprintf(periodic_command, "%s", argv[2]);	
	}
	if (argc == 4) {
		period_time = atoi(argv[3]);
		if (period_time<=0) period_time  = PERIOD_DEFAULT;
	}	
	
	(void) signal(SIGTERM, handle_sigterm);
	(void) signal(SIGPIPE, handle_sigterm);

	do_daemon();
	save_pid(portnum);	
	
	do_filter();
	close_fifos();  

	return 0;
}

//
// close all fifos
//
void close_fifos()
{
	close(s2f_fd);
	close(f2s_fd);
	close(e2f_fd);
	close(f2e_fd);
}

//
// read on serial port and write the received data to socket.
//
void *s2e_thread (void *arg)
{
	char buffer[BUFFER_SIZE];
	int nread=0, nwrite=0, length=0, offset=0 ;
	char fifoname[128];

	// enable cancelation state and set cancelation type to asynchronous 	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, STOF_FIFO, portnum);
	s2f_fd = open(fifoname, O_RDONLY);
	if (s2f_fd < 0) pthread_exit(NULL);
	
	sprintf(fifoname, FTOE_FIFO, portnum);
	f2e_fd = open(fifoname, O_WRONLY);
	if (f2e_fd < 0) pthread_exit(NULL);
	
	while(!bexit) {
		if (!length) {	
			// read on serial port
			nread = read(s2f_fd, buffer, sizeof(buffer));
			if (nread<=0) continue;
			length = nread, offset = 0;
		}
		if (length) {
			// write the all received data to socket
			nwrite = write(f2e_fd, buffer+offset, length);
			if (nwrite>0) {
				length -= nwrite, offset += nwrite;	
			}
		}
	}
	pthread_exit(NULL);
}


//
// read on socket and write the received data to serial port.
//
void *e2s_thread (void *arg)
{
	char buffer[BUFFER_SIZE];
	int nread=0, nwrite=0, length=0, offset=0 ;
	char fifoname[128];

	//
	// enable cancelation state and set cancelation type to asynchronous 	
	//	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	sprintf(fifoname, ETOF_FIFO, portnum);
	e2f_fd = open(fifoname, O_RDONLY);
	if (e2f_fd < 0) pthread_exit(NULL);
	
	sprintf(fifoname, FTOS_FIFO, portnum);
	f2s_fd = open(fifoname, O_WRONLY);
	if (f2s_fd < 0) pthread_exit(NULL);
	
	while(!bexit) {
		if (!length) {		
			// read on socket
			nread = read(e2f_fd, buffer, sizeof(buffer));
			if (nread<=0) continue;
			length = nread, offset = 0;
		}
		if (length) {
			// write the all received data to serial port
			nwrite = write(f2s_fd, buffer+offset, length);
			if (nwrite>0) {
				length -= nwrite, offset += nwrite;	
			}
		}
	}
	pthread_exit(NULL);
}

//
//	Every period_time seconds, write specified command string to serial port.
//
void *periodic_thread (void *arg)
{
	char buffer[1024];

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// if f2s_fd is not ready, sleep
	while (f2s_fd < 0) sleep(2);

	sprintf(buffer, "%s\r\n", periodic_command);
	
	while(!bexit) {
		sleep(period_time);
		pthread_mutex_lock(&serial_write_mutex);
		write(f2s_fd, buffer, strlen(buffer));
		pthread_mutex_unlock(&serial_write_mutex);
	}
	pthread_exit(NULL);
}


void do_filter()
{
	int result;
	pthread_t s2e_th, e2s_th, periodic_th;
	void *thread_result;

	result = pthread_mutex_init(&serial_write_mutex, NULL);
	if (result) return ;

	// create threads	
	result = pthread_create(&s2e_th, NULL, s2e_thread, NULL);
	result = pthread_create(&e2s_th, NULL, e2s_thread, NULL);
	result = pthread_create(&periodic_th, NULL, periodic_thread, NULL);

	// wait until receives SIGTERM
	while(!bexit) pause();
	
	// cancel all threads	
	pthread_cancel(s2e_th);
	pthread_cancel(e2s_th);
	pthread_cancel(periodic_th);
	
	// wait until all threads is canceled.	
	pthread_join(s2e_th, &thread_result);
	pthread_join(e2s_th, &thread_result);
	pthread_join(periodic_th, &thread_result);
	
	pthread_mutex_destroy(&serial_write_mutex);
}


